# coding=utf-8

from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.ray import Ray3D

import unittest
import pytest
import math


class Ray3DTestCase(unittest.TestCase):
    """Test for Ray3D"""

    def test_ray3d_init(self):
        """Test the initalization of Ray3D objects and basic properties."""
        pt = Point3D(2, 0, 2)
        vec = Vector3D(0, 2, 0)
        ray = Ray3D(pt, vec)
        str(ray)  # test the string representation of the ray

        assert ray.p == Point3D(2, 0, 2)
        assert ray.v == Vector3D(0, 2, 0)

        flip_ray = ray.reverse()
        assert flip_ray.p == Point3D(2, 0, 2)
        assert flip_ray.v == Vector3D(0, -2, 0)

    def test_ray3_to_from_dict(self):
        """Test the to/from dict of Ray3D objects."""
        pt = Point3D(2, 0, 2)
        vec = Vector3D(0, 2, 0)
        ray = Ray3D(pt, vec)
        ray_dict = ray.to_dict()
        new_ray = Ray3D.from_dict(ray_dict)
        assert isinstance(new_ray, Ray3D)
        assert new_ray.to_dict() == ray_dict

    def test_linerayment2_mutability(self):
        """Test the immutability of Ray3D objects."""
        pt = Point3D(2, 0, 0)
        vec = Vector3D(0, 2, 0)
        ray = Ray3D(pt, vec)

        assert isinstance(ray, Ray3D)
        with pytest.raises(AttributeError):
            ray.p.x = 3
        with pytest.raises(AttributeError):
            ray.v.x = 3
        with pytest.raises(AttributeError):
            ray.p = Point3D(0, 0, 0)
        with pytest.raises(AttributeError):
            ray.v = Vector3D(2, 2, 0)

        ray_copy = ray.duplicate()
        assert ray.p == ray_copy.p
        assert ray.v == ray_copy.v

    def test_move(self):
        """Test the Ray3D move method."""
        pt = Point3D(2, 0, 2)
        vec = Vector3D(0, 2, 0)
        ray = Ray3D(pt, vec)

        vec_1 = Vector3D(2, 2, 2)
        new_ray = ray.move(vec_1)
        assert new_ray.p == Point3D(4, 2, 4)
        assert new_ray.v == vec
        assert new_ray.v.magnitude == 2

    def test_scale(self):
        """Test the Ray3D scale method."""
        pt = Point3D(2, 2, 2)
        vec = Vector3D(0, 2, 0)
        ray = Ray3D(pt, vec)

        origin_1 = Point3D(0, 2, 2)
        origin_2 = Point3D(1, 1, 2)
        new_ray = ray.scale(2, origin_1)
        assert new_ray.p == Point3D(4, 2, 2)
        assert new_ray.v == Point3D(0, 4, 0)
        assert new_ray.v.magnitude == 4

        new_ray = ray.scale(2, origin_2)
        assert new_ray.p == Point3D(3, 3, 2)
        assert new_ray.v == Point3D(0, 4, 0)

    def test_scale_world_origin(self):
        """Test the Ray3D scale method with None origin."""
        pt = Point3D(2, 2, 2)
        vec = Vector3D(0, 2, 0)
        ray = Ray3D(pt, vec)

        new_ray = ray.scale(2)
        assert new_ray.p == Point3D(4, 4, 4)
        assert new_ray.v == Point3D(0, 4)
        assert new_ray.v.magnitude == 4

    def test_rotate(self):
        """Test the Ray3D rotate method."""
        pt = Point3D(2, 2, 2)
        vec = Vector3D(0, 2, 0)
        ray = Ray3D(pt, vec)
        origin_1 = Point3D(0, 0, 0)
        axis_1 = Vector3D(1, 0, 0)

        test_1 = ray.rotate(axis_1, math.pi, origin_1)
        assert test_1.p.x == pytest.approx(2, rel=1e-3)
        assert test_1.p.y == pytest.approx(-2, rel=1e-3)
        assert test_1.p.z == pytest.approx(-2, rel=1e-3)
        assert test_1.v.x == pytest.approx(0, rel=1e-3)
        assert test_1.v.y == pytest.approx(-2, rel=1e-3)
        assert test_1.v.z == pytest.approx(0, rel=1e-3)

        test_2 = ray.rotate(axis_1, math.pi/2, origin_1)
        assert test_2.p.x == pytest.approx(2, rel=1e-3)
        assert test_2.p.y == pytest.approx(-2, rel=1e-3)
        assert test_2.p.z == pytest.approx(2, rel=1e-3)
        assert test_2.v.x == pytest.approx(0, rel=1e-3)
        assert test_2.v.y == pytest.approx(0, rel=1e-3)
        assert test_2.v.z == pytest.approx(2, rel=1e-3)

    def test_rotate_xy(self):
        """Test the Ray3D rotate_xy method."""
        pt = Point3D(2, 2, 2)
        vec = Vector3D(0, 2, 0)
        ray = Ray3D(pt, vec)
        origin_1 = Point3D(0, 2, 2)

        test_1 = ray.rotate_xy(math.pi, origin_1)
        assert test_1.p.x == pytest.approx(-2, rel=1e-3)
        assert test_1.p.y == pytest.approx(2, rel=1e-3)
        assert test_1.v.x == pytest.approx(0, rel=1e-3)
        assert test_1.v.y == pytest.approx(-2, rel=1e-3)

        test_2 = ray.rotate_xy(math.pi/2, origin_1)
        assert test_2.p.x == pytest.approx(0, rel=1e-3)
        assert test_2.p.y == pytest.approx(4, rel=1e-3)
        assert test_2.v.x == pytest.approx(-2, rel=1e-3)
        assert test_2.v.y == pytest.approx(0, rel=1e-3)

    def test_reflect(self):
        """Test the Ray3D reflect method."""
        pt = Point3D(2, 2, 2)
        vec = Vector3D(0, 2, 0)
        ray = Ray3D(pt, vec)

        origin_1 = Point3D(0, 1, 2)
        origin_2 = Point3D(1, 1, 2)
        normal_1 = Vector3D(0, 1, 0)
        normal_2 = Vector3D(-1, 1, 0).normalize()

        assert ray.reflect(normal_1, origin_1).p == Point3D(2, 0, 2)
        assert ray.reflect(normal_1, origin_1).v == Vector3D(0, -2, 0)
        assert ray.reflect(normal_1, origin_2).p == Point3D(2, 0, 2)
        assert ray.reflect(normal_1, origin_2).v == Vector3D(0, -2, 0)

        test_1 = ray.reflect(normal_2, origin_2)
        assert test_1.p == Point3D(2, 2, 2)
        assert test_1.v.x == pytest.approx(2, rel=1e-3)
        assert test_1.v.y == pytest.approx(0, rel=1e-3)
        assert test_1.v.z == pytest.approx(0, rel=1e-3)

        test_2 = ray.reflect(normal_2, origin_1)
        assert test_2.p.x == pytest.approx(1, rel=1e-3)
        assert test_2.p.y == pytest.approx(3, rel=1e-3)
        assert test_2.p.z == pytest.approx(2, rel=1e-3)
        assert test_2.v.x == pytest.approx(2, rel=1e-3)
        assert test_2.v.y == pytest.approx(0, rel=1e-3)
        assert test_2.v.z == pytest.approx(0, rel=1e-3)

    def test_closest_point(self):
        """Test the Ray3D closest_point method."""
        pt = Point3D(2, 2, 2)
        vec = Vector3D(0, 2, 0)
        ray = Ray3D(pt, vec)

        near_pt = Point3D(3, 3, 0)
        assert ray.closest_point(near_pt) == Point3D(2, 3, 2)
        near_pt = Point3D(2, 0, 0)
        assert ray.closest_point(near_pt) == Point3D(2, 2, 2)
        near_pt = Point3D(1, 5, 0)
        assert ray.closest_point(near_pt) == Point3D(2, 5, 2)

    def test_distance_to_point(self):
        """Test the Ray3D distance_to_point method."""
        pt = Point3D(2, 2, 2)
        vec = Vector3D(0, 2, 0)
        ray = Ray3D(pt, vec)

        near_pt = Point3D(3, 3, 2)
        assert ray.distance_to_point(near_pt) == 1
        near_pt = Point3D(2, 0, 2)
        assert ray.distance_to_point(near_pt) == 2
        near_pt = Point3D(1, 5, 2)
        assert ray.distance_to_point(near_pt) == 1


if __name__ == "__main__":
    unittest.main()
